use actix_cors::Cors;
use actix_web::{
    self,
    error::InternalError,
    get,
    http::StatusCode,
    middleware::Logger,
    rt::{self, net::TcpStream},
    web::{self, Data},
    App, Error, HttpRequest, HttpResponse, HttpServer, Responder,
};
use actix_ws::{AggregatedMessage, CloseCode, Closed};
use anyhow::Context;
use futures_util::StreamExt;
use rustls::pki_types::CertificateDer;
use serde_json::json;
use std::{
    env,
    fs::{self, File},
    io::{self, BufReader},
    num::ParseIntError,
};
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use wildmatch::WildMatch;

pub struct TlsConfig {
    pub private_key_file: String,
    pub certificate_chain_file: String,
}

pub struct ServerConfig {
    pub upstream_mud: String,
    pub banner_to_mud: String,
    pub listen_port: u16,
    pub bind_address: String,
    pub startup_script_file: String,
    pub tls_config: Option<TlsConfig>,
    pub allowed_origins: Vec<WildMatch>,
}

#[get("/ws")]
async fn ws(
    config_data: Data<ServerConfig>,
    req: HttpRequest,
    body: web::Payload,
) -> impl Responder {
    let (response, mut session, stream) = actix_ws::handle(&req, body)?;
    let mut stream = stream.aggregate_continuations().max_continuation_size(1024);

    let mut tcp_stream: TcpStream = TcpStream::connect(&config_data.upstream_mud).await?;
    let subst_banner = config_data
        .banner_to_mud
        .replace(
            "%i",
            &req.peer_addr()
                .map(|a| a.ip().to_string())
                .unwrap_or_else(|| "unknown".to_owned()),
        )
        .replace("%n", "\r\n");
    tcp_stream.write_all(subst_banner.as_bytes()).await?;

    let script = fetch_startup_script(&config_data.startup_script_file)
        .map_err(|e| InternalError::new(e, StatusCode::INTERNAL_SERVER_ERROR))?;

    rt::spawn(async move {
        if session
            .text(
                json!({
                    "RunLua": script
                })
                .to_string(),
            )
            .await
            .is_err()
        {
            let _ = session.close(Some(CloseCode::Normal.into())).await;
            return;
        }
        let mut readbuf: [u8; 1024] = [0; 1024];
        loop {
            tokio::select! {
                ws_msg = stream.next() => {
                    match ws_msg {
                        None => break,
                        Some(Err(_e)) => break,
                        Some(Ok(AggregatedMessage::Binary(bin))) => {
                            if tcp_stream.write_all(&bin).await.is_err() {
                                break
                            }
                        }

                        Some(Ok(AggregatedMessage::Ping(msg))) => {
                            if let Err(Closed) = session.pong(&msg).await {
                                break
                            }
                        }
                        Some(Ok(_)) => {},
                    }
                },
                tcp_data_len = tcp_stream.read(&mut readbuf) => {
                    match tcp_data_len {
                        Err(_e) => break,
                        Ok(0) => break,
                        Ok(n) =>
                            if let Err(Closed) = session.binary(readbuf[0..n].to_vec()).await {
                                break;
                            }
                    }

                }
            }
        }
        let _ = session.close(Some(CloseCode::Normal.into())).await;
    });

    Ok::<HttpResponse, Error>(response)
}

fn extract_server_config_from_environment() -> anyhow::Result<ServerConfig> {
    Ok(ServerConfig {
        upstream_mud: env::var("UPSTREAM_MUD").map_err(|_| {
            anyhow::Error::msg(
                "Expected UPSTREAM_MUD environment variable specifying where to connect.",
            )
        })?,
        banner_to_mud: env::var("BANNER_TO_MUD").map_err(|_| {
            anyhow::Error::msg(
                "Expected BANNER_TO_MUD environment variable specifying message to send to MUD.",
            )
        })?,
        listen_port: env::var("LISTEN_PORT")
            .map_err(|_| {
                anyhow::Error::msg(
                    "Expected LISTEN_PORT environment variable specifying port to listen on.",
                )
            })
            .and_then(|v| {
                v.parse::<u16>().map_err(|_e: ParseIntError| {
                    anyhow::Error::msg("LISTEN_PORT should be a decimal port number")
                })
            })?,
        bind_address: env::var("BIND_ADDRESS").unwrap_or_else(|_| "::".to_owned()),
        startup_script_file: env::var("STARTUP_SCRIPT_FILE")
            .map_err(|_| {
                anyhow::Error::msg(
                    "Expected STARTUP_SCRIPT_FILE environment variable containing filename of script to send to client.",
                )
            })?,
        tls_config: env::var("PRIVATE_KEY_FILE").ok().map(|pk| Ok::<TlsConfig, anyhow::Error>(TlsConfig {
            private_key_file: pk,
            certificate_chain_file: env::var("CERTIFICATE_CHAIN_FILE").map_err(|_| {
                anyhow::Error::msg(
                    "Expected CERTIFICATE_CHAIN_FILE when PRIVATE_KEY_FILE is set."
                )
            })?,
        })).transpose()?,
        allowed_origins: env::var("ALLOWED_ORIGINS")
            .map_err(|_| {
                anyhow::Error::msg("Expected ALLOWED_ORIGINS with list of origins to accept") })?
            .split(',').map(|v| WildMatch::new(v)).collect()
    })
}

// We load this on each connection so it can change.
fn fetch_startup_script(filename: &str) -> anyhow::Result<String> {
    Ok(fs::read_to_string(filename)?)
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    env_logger::init();

    let config = extract_server_config_from_environment()?;

    let data = Data::new(config);
    let config = data.get_ref();

    // Do this early so we fail fast if config is wrong.
    fetch_startup_script(&config.startup_script_file)
        .context("While checking STARTUP_SCRIPT_FILE can be read")?;

    let server_data = data.clone();
    let server = HttpServer::new(move || {
        let logger = Logger::default();
        let cors_server_data = server_data.clone();
        let cors = Cors::default().allowed_origin_fn(move |origin, _| {
            cors_server_data
                .allowed_origins
                .iter()
                .any(|o| o.matches(origin.to_str().unwrap_or("invalid")))
        });
        App::new()
            .wrap(logger)
            .wrap(cors)
            .app_data(server_data.clone())
            .service(ws)
    });

    let server = match &config.tls_config {
        None => server.bind((config.bind_address.clone(), config.listen_port))?,
        Some(tls_config) => {
            let certs: Vec<CertificateDer> = rustls_pemfile::certs(&mut BufReader::new(
                &mut File::open(&tls_config.certificate_chain_file)?,
            ))
            .collect::<Result<Vec<CertificateDer>, io::Error>>()?;
            let key = rustls_pemfile::private_key(&mut BufReader::new(&mut File::open(
                &tls_config.private_key_file,
            )?))?
            .expect("No private key found in private key file");
            let tls_config = rustls::server::ServerConfig::builder()
                .with_no_client_auth()
                .with_single_cert(certs, key)
                .expect("Bad certificate/key");
            server.bind_rustls_0_23(
                (config.bind_address.clone(), config.listen_port),
                tls_config,
            )?
        }
    };
    server.run().await?;
    Ok(())
}
